/*************************************************************
 *                 bbomberman                                *
 *  (c) 2004 by Bernhard Eder (bb-generation)                *
 *  http://bbgen.net/bomberman for updates                   *
 *  Webmaster@bbgen.net for bug reports and other stuff      *
 * --------------------------------------------------------- *
 * This program is free software; you can modify it under    *
 * the terms of the GNU General Public License as published  *
 * by the Free Software Foundation, located in LICENSE       *
 *                                                           *
 *************************************************************/
 
#include "cl_game.h"
#include "cl_server_net.h"

/* used by implementation */
#include <iostream>
#include <sstream>
#include <ctime>
#include <cmath>


/*
 * cl_player implementation
 */

/* copy constructor */
inline cl_player::cl_player(const cl_player& player) :
  game(player.game), keyset(player.keyset), num(player.num)
{
  bomberman = player.bomberman;
}

/* constructor */
inline cl_player::cl_player(cl_game& game, const unsigned short& keyset, const unsigned short& num) :
  game(game), keyset(keyset), num(num)
{
}

/* assignment operator overload */
inline cl_player& cl_player::operator= (const cl_player& player)
{
  return *(new cl_player(player));
}


/*
 * cl_game implementation
 */

/* constructor */
cl_game::cl_game(const cl_init& Inital) :
  Inital(Inital)
{
  /* initialize pseudo-random number generator */
  std::srand(std::time(NULL));
  
  /* initialize contained objects in specific order */
  //Inital = new cl_init;
  BB_DISPLAYWIDTH = Inital.iDspWidth;
  Images = new cl_images;
  Settings = new cl_settings;
}

/* destructor */
cl_game::~cl_game()
{
  /* delete objects created manually in constructor */
  delete Settings;
  delete Images;
  //delete Inital;
}

void cl_game::DrawBField()
{ 
  SDL_Rect bdrect;
  
  for (int i = 0; i < server->map_fields_y(); i++)  //rows
    for (int j = 0; j < server->map_fields_x(); j++)  //columns
    {
      bdrect.x = 24 * j;
      bdrect.y = 24 * i;

      /* draw map base field at this position */
      switch (server->map(j, i))
      {
        case BB_MAP_FIELD_EMPTY:
          SDL_BlitSurface(Images->gras,NULL,BField,&bdrect);
          break;
        case BB_MAP_FIELD_BOX:
          SDL_BlitSurface(Images->box[0],NULL,BField,&bdrect);
          break;
        case BB_MAP_FIELD_BOX_FIXED:
          SDL_BlitSurface(Images->box[1],NULL,BField,&bdrect);
      }

      /* FIXME */
#if 0
      /* draw special fields at this position, if any */
      cl_server::dynamic_fields_type  dynamic_fields = server->dynamic_fields(j, i);
      for (cl_server::dynamic_fields_type::iterator dynamic_field = dynamic_fields.begin();
           dynamic_field != dynamic_fields.end(); dynamic_field++)
      {
        switch (dynamic_field->second)
        {
          case BB_DYN_FIELD_PUP_BOMBSMAX:
            //SDL_BlitSurface(Images->gras,NULL,BField,&bdrect);
            SDL_BlitSurface(Images->powerup[0], NULL, BField, &bdrect);
            break;
          case BB_DYN_FIELD_PUP_EXPFORCE:
            //SDL_BlitSurface(Images->gras,NULL,BField,&bdrect);
            SDL_BlitSurface(Images->powerup[1], NULL, BField, &bdrect);
            break;
          case BB_DYN_FIELD_EXPLOSION:
            /* FIXME: UI */
            SDL_BlitSurface(Images->explosion, NULL, BField, &bdrect); 
            break;
        }
      }
#endif
    }

  DrawBombs();

  /* draw special fields */
  cl_server::dynamic_fields_type  dynamic_fields = server->dynamic_fields();
  for (cl_server::dynamic_fields_type::const_iterator  dynamic_field = dynamic_fields.begin();
      dynamic_field != dynamic_fields.end(); dynamic_field++)
  {
    bdrect.x = 24 * dynamic_field->first.x;
    bdrect.y = 24 * dynamic_field->first.y;

    switch (dynamic_field->second)
    {
      case BB_DYN_FIELD_PUP_BOMBSMAX:
        //SDL_BlitSurface(Images->gras,NULL,BField,&bdrect);
        SDL_BlitSurface(Images->powerup[0], NULL, BField, &bdrect);
        break;
      case BB_DYN_FIELD_PUP_EXPFORCE:
        //SDL_BlitSurface(Images->gras,NULL,BField,&bdrect);
        SDL_BlitSurface(Images->powerup[1], NULL, BField, &bdrect);
        break;
      case BB_DYN_FIELD_EXPLOSION:
        /* FIXME: UI */
        SDL_BlitSurface(Images->explosion, NULL, BField, &bdrect); 
        break;
    }
  }

  DrawPlBombermans();
  
  SDL_Rect drect;
  drect.x = (Inital.iX - server->map_fields_x() * 24 - BB_DISPLAYWIDTH) / 2;
  drect.y = (Inital.iY - server->map_fields_y() * 24) / 2;
  SDL_BlitSurface(BField,NULL,Inital.screen,&drect);
}

//Draw the bombs and checks if any bomb have to be "killed" >:D
void cl_game::DrawBombs()
{
  for (cl_server::bombermans_type::iterator bman = server->bombermans.begin();
       bman < server->bombermans.end(); bman++)
  {
    for (cl_bomberman::bombs_type::iterator bomb = (*bman)->bombs.begin();
         bomb < (*bman)->bombs.end(); bomb++)
    {
      SDL_Rect drect;
      drect.x = 24 * (*bomb)->x;
      drect.y = 24 * (*bomb)->y;

      if ((*bomb)->status() < BB_BOMB_STAT_EXPLODING)
        SDL_BlitSurface(Images->bomb[(*bomb)->status()], NULL, BField, &drect);
    }
  }
}

// Action: 0down 1up 2right 3left
// reality 0left 1right 2up 3down

void cl_game::DrawBomberman(const cl_bomberman& bman)
{
  SDL_Rect srcrect,bdrect;

#if 0
  /* FIXME: UI; inversed direction! */
  unsigned short action = bman.direction;
  if(action%2 == 0) srcrect.x=0;
  else srcrect.x=25;
  if(action <= 1) srcrect.y=0;
  else srcrect.y=25;
#endif

  switch (bman.direction)
  {
    case BB_DIRECTION_LEFT:
      srcrect.x = 25; srcrect.y = 25; break;
    case BB_DIRECTION_RIGHT:
      srcrect.x =  0; srcrect.y = 25; break;
    case BB_DIRECTION_UP:
      srcrect.x = 25; srcrect.y =  0; break;
    case BB_DIRECTION_DOWN:
      srcrect.x =  0; srcrect.y =  0; break;
    default:
      srcrect.x =  0; srcrect.y =  0; break;
  }

  srcrect.h = 24;
  srcrect.w = 24;
  bdrect.x = 24 * bman.x;
  bdrect.y = 24 * bman.y;
  
  /* FIXME: server; this is slightly different from bb's last version;
   *   index into bomberman array of images is server's bomberman object's
   *   id, not player's numer -- it might belong to a remote player, which
   *   doesn't have any player object on this host
   */
  SDL_BlitSurface(Images->bomberman[bman.id], &srcrect, BField, &bdrect);  
}

void cl_game::DrawPlBombermans()
{
  for (cl_server::bombermans_type::const_iterator bman = server->bombermans.begin();
       bman < server->bombermans.end(); bman++)
  {
    if ((*bman)->alive)
      DrawBomberman(**bman);
  }
}

void cl_game::DrawDisplay(const int& frlength)
{
  FillSurface(Display,Images->texture[0]);
  SDL_Rect drect,trect;
  drect.x = Inital.screen->w - BB_DISPLAYWIDTH;
  drect.y = 0;

  trect.x = 35;
  
  SDL_Color clr = { 128,128,180, 0 };
  std::stringstream strstream;

//Player display...
  for (players_type::iterator player = players.begin();
       player < players.end(); player++)
  {
    trect.y = 30 * (*player)->num + 8;
    strstream.str("");
    strstream << (*player)->bomberman->name;
    text_surface=TTF_RenderText_Solid(display_font,strstream.str().c_str(),clr);
    SDL_BlitSurface(text_surface,NULL,Display,&trect);
    
    SDL_Rect srcrect,bdrect;
    srcrect.x=0;
    srcrect.y=0;
    srcrect.h=24;
    srcrect.w=24;
    bdrect.x=5;
    bdrect.y= 30 * (*player)->num + 6;
    
    
    SDL_BlitSurface(Images->bomberman[(*player)->num],&srcrect,Display,&bdrect);  
    if (!(*player)->bomberman->alive)
      SDL_BlitSurface(Images->cross,NULL,Display,&bdrect);
    SDL_FreeSurface(text_surface);
    
//How many bombs & explosion force?
#if 0
    if (player->keyset < 10)
    {
#endif
//bombs
      strstream.str("");
      strstream << (*player)->bomberman->bombs_max;
      text_surface = TTF_RenderText_Solid(display_font,strstream.str().c_str(),clr);
      bdrect.x = BB_DISPLAYWIDTH - 60;
      SDL_BlitSurface(Images->bomb[0],NULL,Display,&bdrect);
      
      bdrect.x += 6;
      bdrect.y += 6;
      SDL_BlitSurface(text_surface,NULL,Display,&bdrect);
      SDL_FreeSurface(text_surface);

//explosion force
      strstream.str("");
      strstream << (*player)->bomberman->expforce;
      text_surface = TTF_RenderText_Solid(display_font,strstream.str().c_str(),clr);
      bdrect.x = BB_DISPLAYWIDTH - 31;
      bdrect.y -= 3;
      SDL_BlitSurface(Images->explosion,NULL,Display,&bdrect);
      bdrect.y += 3;
      bdrect.x += 7;
      SDL_BlitSurface(text_surface,NULL,Display,&bdrect);
      SDL_FreeSurface(text_surface);
      
#if 0
    }
#endif
    
  }

//Frame Counter..
  strstream.str("");
  strstream << "Frame length: " << frlength;
  text_surface=TTF_RenderText_Solid(display_font,strstream.str().c_str(),clr);
  trect.x = 22;
  trect.y = 30 * server->bombermans.size() + 20;
  
  SDL_BlitSurface(text_surface,NULL,Display,&trect);
  
  SDL_FreeSurface(text_surface);

//  text_surface=TTF_RenderText_Solid(display_font,strstream.str().c_str(),clrBlack);

//  SDL_BlitSurface(text_surface,NULL,Display,NULL);
    
  
  
  SDL_BlitSurface(Display,NULL,Inital.screen,&drect);
}

void cl_game::FillSurface(SDL_Surface* screen, SDL_Surface* texture)
{
//  SDL_BlitSurface(texture, NULL, screen, NULL);
  SDL_Rect srcrect, dstrect;
  srcrect.h = texture->h;
  for(int i=0;i<=ceil(screen->h/texture->h);i++) //rows
    for(int j=0;j<=ceil(screen->w/texture->w);j++) //collumns
    {
      dstrect.x = texture->w*j;
      dstrect.y = texture->h*i;
      SDL_BlitSurface(texture,NULL,screen,&dstrect);
    }
}

/* FIXME: whatsthis; bb..? */
//0 0    1 21   2 0   3 21
//0 0    1 0    2 21  3 21

/* initialize player objects */
void cl_game::init_players()
{
  /* reset container */
  players.clear();

  /* create one player object for each player */
  for (unsigned short i = 0; i < Inital.playercount; i++)
  {
    player_ptr  player(new cl_player(*this, i, i));
    player->bomberman = server->bombermans[i].get();
    players.push_back(player);
  }
}

/* parse SDL key event (polled in event loop) */
bool cl_game::parse_key(const SDLKey& key)
{
  /* iterate through all local players */
  for (players_type::iterator player = players.begin();
       player < players.end(); player++)
  {
    /* skip unless alive */
    if (!(*player)->bomberman->alive)
      continue;

    /* check key event against every key from player's keyset */
    for (unsigned int action = 0; action <= 4; action++)
    {
      if (key == Settings->bk[(*player)->keyset][action])
      {
        /* movement action? */
        if ((0 <= action) && (action <= 3))
          /* move player's bomberman in this direction */
          (*player)->bomberman->move((bb_direction) action);
        /* bomb placement action? */
        else
          (*player)->bomberman->place_bomb();

        /* match found, so don't search for any further one */
        return true;
      }
    }
  }

  /* indicate no matching key found in players' keysets */
  return false;
}

/* play the game */
void cl_game::play(cl_server& _server)
{
  std::cerr << "+ cl_game::play()" << std::endl;
  server = &_server;

  /* wait for server to finish initialization */
  while (!server->is_initialized())
    server->tick();

  /* initialize player objects */
  std::cerr << "* cl_game::play() initializing player objects" << std::endl;
  init_players();
  
  /* initialize UI */
  std::cerr << "* cl_game::play() initializing user interface" << std::endl;
  display_font = TTF_OpenFont("Fixedsys500c.ttf", 16);
  Display = SDL_CreateRGBSurface(SDL_HWSURFACE, 300, Inital.screen->h, 32, 0, 0, 0, 0);
  BField = SDL_CreateRGBSurface(SDL_HWSURFACE, server->map_fields_x() * 24, server->map_fields_y() * 24, 32, 0, 0, 0, 0);
  SDL_EnableKeyRepeat(0, 0);  // 160, 80

#if 0
  /* perform initial screen update */
  /* FIXME: whatsthis; why is this in the opposite order as below..?
   *   perhaps move screen updates to begin of loop -> drop these lines here?
   */
  std::cerr << "* cl_game::play() performing initial screen update" << std::endl;
  DrawDisplay(0);
  DrawBField();
  SDL_Flip(Inital.screen);
#endif
  
  /* enter event loop */
  bool       brun = 1;
  SDL_Event  event;
  long       begtime, curtime;
  while (brun)
  {
    std::cerr << "* cl_game::play() event loop" << std::endl;

    /* remember start of frame */
    begtime = std::clock() / (CLOCKS_PER_SEC / 1000);  // IMPORTANT! because under linux it's often microseconds

    /* actually perform actions (player movements, bomb explosions, ...) */
    server->tick();

    /* process messages the server has sent to us */
    _M_process_server_events();

    /* update screen */
//    SDL_FillRect(Inital.screen, NULL, 0);
    DrawBField();
    DrawDisplay(std::clock() / (CLOCKS_PER_SEC / 1000) - begtime);
    SDL_Flip(Inital.screen);

    /* process every queued SDL event */
    /* FIXME: this could lead to lockups in case event processing causes new
     *   events to be generated; could also be problematic in networked games,
     *   when a single player stops the server by flooding it with local events
     */
    while (SDL_PollEvent(&event))
    {
      switch (event.type)
      {
        case SDL_KEYDOWN:
          std::cerr << "* cl_game::play() event loop: SDL key event processing" << std::endl;

          /* game abortion key? */
          if (event.key.keysym.sym == SDLK_ESCAPE)
            /* exit game */
            brun = 0;

          /* parse key event for other keys */
          /* FIXME: ui; maybe play rejection sound or something on false..? */
          parse_key(event.key.keysym.sym);
          break;

        case SDL_QUIT:
          /* exit game */
          brun = 0;
      }
    }

    /* calculate frame length; delay further execution if it's too short */
    curtime = std::clock() / (CLOCKS_PER_SEC / 1000);
    if ((curtime - begtime) > 75)
      begtime = curtime - 75;

    /* FIXME: it'll delay only if frame is too short, no matter what
     *   is_delaying() says... or is it better like this? client performance
     *   isn't good at all ATM. :'( (2004-10-12)
     */
    /* only delay if the used server doesn't already do so */
    if (!server->is_delaying())
      SDL_Delay(75 - (curtime - begtime));
//    std::cout << curtime << " " << begtime << " " << curtime - begtime <<  std::endl;
  }
  std::cerr << "Frame Counter stopped at: " << server->tick_count() << "." << std::endl;

  /* deinitialize UI */
  SDL_FreeSurface(Display);
  SDL_FreeSurface(BField);
  /* FIXME: whatthis; what about display_font..? O_o */

  /* deinitialize manually created objects */
  players.clear();

  std::cerr << "- cl_game::play()" << std::endl;
}

/* process messages the server has sent to us */
void cl_game::_M_process_server_events()
{
  cl_proto_queue&  queue = server->clients[0]->to_client;

  while (!queue.empty())
  {
    cl_proto_msg  msg = queue.poll_msg();

    switch (msg.type())
    {
      case BB_PROTO_MSG_EV_BBPUP:
        /* FIXME: sound; play powerup sound */
        break;
      case BB_PROTO_MSG_EV_BBKILL:
        /* FIXME: ui; inform the user or even switch into game-over state */
        /* FIXME: sound; play bomberman-killed sound */
        std::cout << "Player " <<
          server->bombermans[msg.arg_to_int(0)]->name <<
          " killed by Player " <<
          server->bombermans[msg.arg_to_int(1)]->name <<
          std::endl;
        break;

      case BB_PROTO_MSG_DENY_BBMOVE:
        /* FIXME: sound; play move-denied sound */
        break;
      case BB_PROTO_MSG_DENY_BBBOMB:
        /* FIXME: sound; play place-bomb-denied sound */
        break;
      default:
        /* ignore all other messages */
        break;
    }
  }
}

/*
unsigned int cl_game::SFBombs() //Search for bombs
{

}*/
