/*************************************************************
 *                 bbomberman                                *
 *  (c) 2004 by Bernhard Eder (bb-generation)                *
 *  http://bbgen.net/bomberman for updates                   *
 *  Webmaster@bbgen.net for bug reports and other stuff      *
 * --------------------------------------------------------- *
 * This program is free software; you can modify it under    *
 * the terms of the GNU General Public License as published  *
 * by the Free Software Foundation, located in LICENSE       *
 *                                                           *
 *************************************************************/
 
#include "cl_server_remote.h"

/* used in implementation only */
#include <exception>
#include <stdexcept>

/* FIXME: debug */
#include <iostream>


/*
 * cl_server_remote implementation
 */

/* constructor */
cl_server_remote::cl_server_remote(const bomberman_names_type& bomberman_names, const std::string& hostname, const Uint16 port) :
  _M_connected(false), _M_initialized(false)
{
  /* initialize network buffer */
  for (int i = 0; i < BB_NET_BUF_LEN; i++)
    _M_buffer[i] = '\0';

  _M_buffer_insert_ptr = _M_buffer;

  /* initialize network API */
  if (SDLNet_Init())
    throw std::runtime_error("cl_server_remote::cl_server_remote(): can't initialize SDL_net");

  /* resolve host name */
  if (SDLNet_ResolveHost(&_M_host_address, hostname.c_str(), port))
    throw std::runtime_error("cl_server_remote::cl_server_remote(): can't resolve host name");

  /* try connecting to remote server */
  _M_socket = SDLNet_TCP_Open(&_M_host_address);
  if (_M_socket == NULL)
    throw std::runtime_error("cl_server_remote::cl_server_remote(): can't connect to server");

  _M_connected = true;

  /* initialize socket set (for polling data on socket without blocking) */
  _M_socket_set = SDLNet_AllocSocketSet(1);

  if (_M_socket_set == NULL)
    throw std::runtime_error("cl_server_remote::cl_server_remote(): can't allocate socket set");

  SDLNet_TCP_AddSocket(_M_socket_set, _M_socket);

  /* create new client object for communication with cl_game */
  clients.push_back(client_ptr(new cl_client(*this)));

  /* FIXME: get server map size[; get server map] */
}

/* destructor */
cl_server_remote::~cl_server_remote()
{
  /* still connected to real server? */
  if (_M_connected)
  {
    /* close network connection to real server */
    SDLNet_TCP_Close(_M_socket);
    _M_connected = false;
  }

  /* free socket set */
  SDLNet_FreeSocketSet(_M_socket_set);

  /* cleanup network API */
  SDLNet_Quit();
}

/* FIXME: has to be implemented because cl_server_local-specific cl_bomb
 *   requires it... => split cl_bomb into cl_bomb and cl_bomb_local
 *   (or something...)
 */
void cl_server_remote::explode_field(const int x, const int y, const cl_bomberman* causer)
{
  throw std::runtime_error("cl_server_remote::explode_field(): shall never be called, as this is not a real server");
}

bool cl_server_remote::is_initialized() const
{
  return _M_initialized;
}

void cl_server_remote::tick()
{
  /* FIXME: exception handling */
#if 0
    try
    {
      /* let client object pass its messages on to real client via network */
      client_net->pass_proto_msgs();
    }
    catch (std::runtime_error err)
    {
      /* assume network error -- detach client */
      client_net->detach();
      clients.erase(client);
    }
#endif
  /* pass protocol messages on to server */
  _M_pass_proto_msgs();

  /* poll network socket */
  switch (SDLNet_CheckSockets(_M_socket_set, BB_NET_DELAY))
  {
    case -1:
      /* error in library function */
      throw std::runtime_error("cl_server_remote::tick(): unknown network error");

    case 0:
      /* no network activity */
      break;

    default:
  /* FIXME: exception handling */
#if 0
    try
    {
      /* process network events sent from server */
      _M_process_net_events();
    }
    catch (std::runtime_error err)
    {
      /* FIXME: implement */
      /* assume network error -- detach server */
      clients[0]->detach();
      clients.pop_back();
    }
#endif
    /* process network events sent from server */
    _M_process_net_events();

    /* process received protocol messages */
    _M_process_server_msgs();
  }
}

/* pass queued protocol messages on to real server via network */
void cl_server_remote::_M_pass_proto_msgs()
{
  cl_proto_queue&  queue = clients[0]->to_server;

  while (!queue.empty())
  {
    cl_proto_msg  msg    = queue.poll_msg();
    std::string   data   = msg.encode();
    data += '\n';

    /* FIXME: cast; explicit typecast from (const void*) to (void*); sadly, SDL_net API isn't const-ified. :'(
     *   assuming SDLNet_TCP_Send() won't modify its send buffer
     */
    int  length = data.length();
    int  sent = SDLNet_TCP_Send(_M_socket, (void*) data.data(), length);

    if (sent < length)
      throw std::runtime_error("cl_server_remote::_M_pass_proto_msgs(): unknown network error");
  }
}

/* process net events for this client */
void cl_server_remote::_M_process_net_events()
{
  /* process incoming data */
  do
  {
    /* read BB_NET_BUF_LEN bytes from network */
    int size = SDLNet_TCP_Recv(_M_socket, _M_buffer_insert_ptr, BB_NET_BUF_LEN - (_M_buffer_insert_ptr - _M_buffer));
    char*  buffer_begin_ptr = _M_buffer;
    char*  buffer_end_ptr   = _M_buffer_insert_ptr + size;

    /* check for receive error */
    if (size <= 0)
      throw std::runtime_error("cl_server_remote::_M_process_net_events(): unknown network error");

    /* process received data */
    for (char* buffer_ptr = _M_buffer_insert_ptr;
         buffer_ptr < buffer_end_ptr;
         buffer_ptr++)
    {
      /* end-of-line encountered? */
      if (*buffer_ptr == '\n')
      {
        /* convert line to NUL-terminated C string */
        *buffer_ptr = '\0';

        /* convert C string to std::string & enqueue in message queue */
        //_M_queue.queue_msg(cl_proto_msg(std::string(buffer_begin_ptr)));
        std::string  line(buffer_begin_ptr);
        std::cerr << "* cl_server_remote::_M_process_net_events(): received: \"" <<
          line << "\"" << std::endl;
        _M_queue.queue_msg(cl_proto_msg(line));

        /* begin new line after NUL C string terminator */
        buffer_begin_ptr = buffer_ptr + 1;
      }
    }

    /* at least one complete line found in buffer? */
    if (buffer_begin_ptr != _M_buffer)
    {
      /* move trailing chars to begin of buffer */
      for (char* buffer_ptr = buffer_begin_ptr;
           buffer_ptr < buffer_end_ptr;
           buffer_ptr++)
      {
        _M_buffer[buffer_ptr - buffer_begin_ptr] = *buffer_ptr;
      }
    }
    else
    {
      /* check for line length exceeded */
      if ((buffer_end_ptr - buffer_begin_ptr) == BB_NET_BUF_LEN)
        throw std::runtime_error("cl_server_remote::_M_process_net_events(): maximum line length exceeded");
    }

    /* calculate new insert pointer */
    _M_buffer_insert_ptr = _M_buffer + (buffer_end_ptr - buffer_begin_ptr);
  }
  while (SDLNet_SocketReady(_M_socket));
}

/* process protocol messages received from server */
void cl_server_remote::_M_process_server_msgs()
{
  while (!_M_queue.empty())
  {
    cl_proto_msg   msg = _M_queue.poll_msg();
    cl_bomberman*  bman;

    /* FIXME: see below */
#if 0
    try
    {
#endif
      /* examine what action to take */
      switch (msg.type())
      {
        case BB_PROTO_MSG_DENY_BBMOVE:
          /* FIXME: sound; play appropriate sound sample */
          /* adjust indicated bomberman's direction */
          bman = &bomberman(msg.arg_to_int(0));

          /* move should fail (-> DENY) => desync if it succeeds */
          if (bman->do_move((bb_direction) msg.arg_to_int(1)))
            throw std::runtime_error("cl_server_remote::_M_process_server_msgs(): desync");

          break;

        case BB_PROTO_MSG_DENY_BBBOMB:
          /* FIXME: sound; play appropriate sound sample */
          break;

        case BB_PROTO_MSG_EV_BBMOVE:
          /* move indicated bomberman */
          bman = &bomberman(msg.arg_to_int(0));

          /* move should succeed (-> EV) => desync if it fails */
          if (!bman->do_move((bb_direction) msg.arg_to_int(1)))
            throw std::runtime_error("cl_server_remote::_M_process_server_msgs(): desync");

          break;

        case BB_PROTO_MSG_EV_BBBOMB:
          bman = &bomberman(msg.arg_to_int(0));

          if (!bman->do_place_bomb())
            throw std::runtime_error("cl_server_remote::_M_process_server_msgs(): desync");

          break;

        case BB_PROTO_MSG_EV_BBPUP:
          bman = &bomberman(msg.arg_to_int(0));
          bman->do_powerup((bb_dynamic_field) msg.arg_to_int(1));
          break;

        case BB_PROTO_MSG_EV_BBKILL:
          bman = &bomberman(msg.arg_to_int(0));

          /* kill local copy of bomberman */
          bman->do_kill(&bomberman(msg.arg_to_int(1)));

          /* pass server message on to cl_game */
          clients[0]->to_client.queue_msg(msg);
          break;

        case BB_PROTO_MSG_EV_MAPFIELD:
          /* change map field in local map */
          _M_map[msg.arg_to_int(0)][msg.arg_to_int(1)] = (bb_map_field) msg.arg_to_int(2);
          break;

        case BB_PROTO_MSG_EV_DFCREATE:
          /* create dynamic field locally */
          _M_create_dynamic_field(msg.arg_to_int(0), msg.arg_to_int(1), (bb_dynamic_field) msg.arg_to_int(2));
          break;

        case BB_PROTO_MSG_EV_DFERASE:
          /* erase dynamic field locally */
          _M_destroy_dynamic_field(msg.arg_to_int(0), msg.arg_to_int(1), (bb_dynamic_field) msg.arg_to_int(2));
          break;

        case BB_PROTO_MSG_EV_TICK:
          /* advance to next tick (should be (last + 1), but clients
           * should not rely on this)
           */
          _M_ticks = msg.arg_to_int(0);
          break;

        case BB_PROTO_MSG_EV_DELBOMB:
          /* delete local copy of indicated bomb */
          bman = &bomberman(msg.arg_to_int(0));
          /* FIXME: add id field to cl_bomb */
          bman->bombs.erase(bman->bombs.begin());
          break;

        /* the following cases shall only be valid during initialization */
        case BB_PROTO_MSG_INIT_MAPCOL:
          /* FIXME: check for initialization done / throw exception if
           * received later..?
           */
          if (_M_initialized)
            throw std::runtime_error("cl_server_remote::_M_process_server_msgs(): already initialized");
          /* add map column to local map (shall only be used during initialization) */
          _M_decode_map_column(msg.arg(0));
          break;

        case BB_PROTO_MSG_INIT_BMAN:
          /* create local copy of bomberman on server */
          if (_M_initialized)
            throw std::runtime_error("cl_server_remote::_M_process_server_msgs(): already initialized");
          bombermans.push_back(bomberman_ptr(new cl_bomberman_remote(*this,
            msg.arg_to_int(0),                 /* id        */
            msg.arg(1),                        /* name      */
            msg.arg_to_int(2),                 /* alive     */
            msg.arg_to_int(3),                 /* x         */
            msg.arg_to_int(4),                 /* y         */
            (bb_direction) msg.arg_to_int(5),  /* direction */
            msg.arg_to_int(6),                 /* bombs_max */
            msg.arg_to_int(7),                 /* expforce  */
            /* FIXME: only bombermans really handeled locally should get a
             *   pointer to the client object
             */
            clients[0].get()
          )));
          break;

        case BB_PROTO_MSG_INIT_BOMB:
          /* create local copy of bomb on server */
          if (_M_initialized)
            throw std::runtime_error("cl_server_remote::_M_process_server_msgs(): already initialized");
          bman = &bomberman(msg.arg_to_int(0));

          bman->bombs.push_back(cl_bomberman::bomb_ptr(new cl_bomb_remote(*bman,
            msg.arg_to_int(1),  /* x             */
            msg.arg_to_int(2),  /* y             */
            /* FIXME: not really int... *COUGH* */
            msg.arg_to_int(3)   /* creation tick */
          )));
          break;

        case BB_PROTO_MSG_INIT_DONE:
          /* finish initialization phase */
          if (_M_initialized)
            throw std::runtime_error("cl_server_remote::_M_process_server_msgs(): already initialized");
          _M_initialized = true;
          break;

        default:
          throw std::runtime_error("cl_server_remote::_M_process_server_msgs(): invalid message type");
      }
    /* FIXME: exception handling; to be implemented... */
#if 0
    }
    catch (std::runtime_error err)
    {
      /* server message caused an exception -- detach from server */
      detach();
      clients.erase(client);
    }
#endif
  }
}
