/*************************************************************
 *                 bbomberman                                *
 *  (c) 2004 by Bernhard Eder (bb-generation)                *
 *  http://bbgen.net/bomberman for updates                   *
 *  Webmaster@bbgen.net for bug reports and other stuff      *
 * --------------------------------------------------------- *
 * This program is free software; you can modify it under    *
 * the terms of the GNU General Public License as published  *
 * by the Free Software Foundation, located in LICENSE       *
 *                                                           *
 *************************************************************/

#include "cl_proto.h"
#include <iostream>
#include <sstream>
#include <stdlib.h>
#include <exception>
#include <stdexcept>

int main()
{
  cl_proto_queue  queue;
  std::string     str;

  std::cout << "testing cl_proto" << std::endl;

  /* throw exception on EOF */
  std::cin.exceptions(std::cin.eofbit);
  try
  {
    /*
     * -> string
     */
    std::cout << "decode from string:" << std::endl;
    while (std::getline(std::cin, str), !str.empty())
    {
      /* decode user input string to message */
      try
      {
        queue.queue_decode_msg(str);
      }
      catch (std::runtime_error err)
      {
        std::cerr << "caught something..: " << err.what() << std::endl;
        continue;
      }

      /* output decoded message parts back to user */
      cl_proto_msg msg = queue.poll_msg();
      std::cout << "type: " << msg.type();
      
      for (cl_proto_msg::args_type::iterator i = msg.args().begin();
           i < msg.args().end(); i++)
        std::cout << "; arg: \"" << *i << "\"";
      std::cout << std::endl;

      std::cout << std::endl;
    }

    /*
     * string ->
     */
    std::cout << "to string:" << std::endl;
    while (std::getline(std::cin, str).eof(), !str.empty())
    {
      std::stringstream        strstream(str);
      int                      type;
      cl_proto_msg::args_type  args;

      /* decode user input string to message on our own */
      strstream >> type;

      while (std::getline(std::cin, str).eof(), !str.empty())
        args.push_back(str);

      /* queue self-decoded message */
      queue.queue_msg(cl_proto_msg((bb_proto_msg) type, args));

      /* output re-encoded message back to user */
      try
      {
        std::cout << queue.poll_encode_msg() << std::endl;
      }
      catch (std::runtime_error err)
      {
        std::cerr << "caught something..: " << err.what() << std::endl;
        continue;
      }

      std::cout << std::endl;
    }
  }
  /* catch EOF */
  catch (std::ios_base::failure err)
  {
    /* really EOF? exit, then */
    if (std::cin.eof() && !std::cin.bad())
      return 1;

    /* some other exception; report to user */
    std::cout << "caught exception: " << err.what() << std::endl;
    return 2;
  }

  std::cout << "all tests completed" << std::endl;

  /* success */
  return 0;
}
